﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

// Note there is a bug in the TreeView control. If you double-click a node,
// the node's appearance changes but its Checked property doesn't change.
// You can fix this by subclassing the TreeView and ignoring WM_LBUTTONDBLCLK messages.
//
//  public class MyTreeView : TreeView
//  {
//      protected override void WndProc(ref Message m)
//      {
//          // Suppress WM_LBUTTONDBLCLK
//          if (m.Msg == 0x203)
//          {
//              m.Result = IntPtr.Zero;
//          }
//          else base.WndProc(ref m);
//      }
//  }
//
// The TreeView also has the annoying habit of displaying each node's text
// in a tooltip if the control isn't wide enough to show it's whole value.

// Add a references to Microsoft.VisualBasic.
using Microsoft.VisualBasic.FileIO;

using System.IO;
using System.Security.Cryptography;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace howto_find_duplicate_files2
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // Save and restore the directory path.
        private void Form1_Load(object sender, EventArgs e)
        {
            txtDirectory.Text = Properties.Settings.Default.Directory;

            splitContainer1.Panel2.BackColor = Color.White;

            // Prepare the preview controls.
            picImage.SizeMode = PictureBoxSizeMode.Zoom;
            picImage.Dock = DockStyle.Fill;
            picImage.Visible = false;
            rchText.Dock = DockStyle.Fill;
            rchText.Visible = false;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            Properties.Settings.Default.Directory = txtDirectory.Text;
            Properties.Settings.Default.Save();
        }

        // Let the user browse for a folder.
        private void btnBrowse_Click(object sender, EventArgs e)
        {
            fbdDirectory.SelectedPath = txtDirectory.Text;
            if (fbdDirectory.ShowDialog() == DialogResult.OK)
                txtDirectory.Text = fbdDirectory.SelectedPath;
        }

        // Search the directory for duplicates.
        private void btnSearch_Click(object sender, EventArgs e)
        {
            Stopwatch watch = new Stopwatch();
            watch.Start();

            Cursor = Cursors.WaitCursor;
            trvFiles.Visible = false;
            trvFiles.Nodes.Clear();
            lblNumDuplicates.Text = "";
            Refresh();

            try
            {
                // Get FileInfos representing the files.
                var get_infos =
                    from string filename in Directory.GetFiles(txtDirectory.Text)
                    select new FileInfo(filename);

                // Group the FileInfos by file length.
                var fileinfo_groups =
                    from FileInfo file_info in get_infos
                    group file_info by file_info.Length into g
                    where g.Count() > 1
                    select g;

                // Flatten the result to get a list of FileInfos.
                var flattened = fileinfo_groups.SelectMany(x => x).ToList();

                // Get a list of the files and their hash values.
                var get_info =
                    from FileInfo file_info in flattened
                    select new
                    {
                        Name = file_info.FullName,
                        Hash = GetHashString(file_info.FullName)
                    };

                // Group the files by hash value.
                var group_infos =
                    from info in get_info
                    group info by info.Hash into g
                    where g.Count() > 1
                    //orderby g.Key
                    select g;

                // Loop through the files.
                int num_groups = 0;
                int num_files = 0;
                foreach (var g in group_infos)
                {
                    num_groups++;
                    TreeNode hash_node = trvFiles.Nodes.Add(g.Key.ToString());
                    foreach (var info in g)
                    {
                        num_files++;
                        TreeNode file_node = new TreeNode(info.Name);
                        file_node.Tag = new FileInfo(info.Name);
                        hash_node.Nodes.Add(file_node);
                    }
                }

                // Display the number of duplicates.
                lblNumDuplicates.Text =
                    (num_files - num_groups).ToString() +
                    " duplicate files";

                // Expand all nodes.
                trvFiles.ExpandAll();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                // Scroll to the top.
                if (trvFiles.Nodes.Count > 0)
                    trvFiles.Nodes[0].EnsureVisible();
                trvFiles.Visible = true;

                Cursor = Cursors.Default;
            }

            watch.Stop();
            Console.WriteLine(watch.Elapsed.TotalSeconds.ToString("0.00") + " seconds");
        }

        // Display the clicked file.
        private void trvFiles_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (Deleting) return;

            // Hide the display controls.
            rchText.Visible = false;
            picImage.Visible = false;
            Refresh();

            // Do nothing for size nodes.
            if (e.Node.Level == 0) return;

            // Get the file's information.
            FileInfo file_info = e.Node.Tag as FileInfo;
            switch (file_info.Extension.ToLower())
            {
                case ".txt":    // Text files.
                case ".html":
                case ".cs":
                case ".csproj":
                case ".resx":
                case ".xml":
                case ".xaml":
                case ".config":
                    rchText.Text = File.ReadAllText(file_info.FullName);
                    rchText.Visible = true;
                    break;

                case ".rtf":    // Rich text.
                    rchText.LoadFile(file_info.FullName);
                    rchText.Visible = true;
                    break;

                case ".jpg":    // Image files.
                case ".jpeg":
                case ".gif":
                case ".png":
                case ".tiff":
                case ".bmp":
                    picImage.Image = LoadBitmapUnlocked(file_info.FullName);
                    picImage.Visible = true;
                    break;

                default:
                    rchText.Text = "Unknown file extension " + file_info.Extension;
                    rchText.Visible = true;
                    break;
            }
        }

        // Select all but the first file in each group.
        private void btnSelectDuplicates_Click(object sender, EventArgs e)
        {
            foreach (TreeNode hash_node in trvFiles.Nodes)
            {
                hash_node.Checked = false;
                hash_node.Nodes[0].Checked = false;
                for (int i = 1; i < hash_node.Nodes.Count; i++)
                    hash_node.Nodes[i].Checked = true;
            }
        }

        // Delete the selected files.
        private bool Deleting = false;
        private void btnDeleteSelected_Click(object sender, EventArgs e)
        {
            Deleting = true;
            trvFiles.Visible = false;
            lblNumDuplicates.Text = "";
            picImage.Visible = false;
            picImage.Image = null;
            rchText.Visible = false;
            Cursor = Cursors.WaitCursor;
            Refresh();

            try
            {
                // Make a list of nodes to delete.
                List<TreeNode> nodes_to_delete = new List<TreeNode>();
                foreach (TreeNode hash_node in trvFiles.Nodes)
                {
                    foreach (TreeNode file_node in hash_node.Nodes)
                    {
                        // See if the file's node is checked.
                        if (file_node.Checked) nodes_to_delete.Add(file_node);
                    }
                }

                // Delete the selected nodes and their files.
                foreach (TreeNode file_node in nodes_to_delete)
                {
                    // Get the FileInfo from the Tag property.
                    FileInfo file_info = file_node.Tag as FileInfo;

                    // Move the file into the recycle bin.
                    DeleteFile(file_info.FullName);

                    // Remove the file from the TreeView.
                    file_node.Remove();
                }

                // Make a list of size nodes with no remaining children.
                nodes_to_delete = new List<TreeNode>();
                foreach (TreeNode hash_node in trvFiles.Nodes)
                {
                    if (hash_node.Nodes.Count == 0) nodes_to_delete.Add(hash_node);
                }

                // Delete the size nodes with no remaining children.
                foreach (TreeNode hash_node in nodes_to_delete)
                {
                    hash_node.Remove();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                // Scroll to the top.
                if (trvFiles.Nodes.Count > 0)
                    trvFiles.Nodes[0].EnsureVisible();
                trvFiles.Visible = true;

                Deleting = false;
                Cursor = Cursors.Default;
            }
        }

        // Delete a file or move it to the recycle bin.
        public static void DeleteFile(string filename)
        {
            DeleteFile(filename, false, false);
        }
        public static void DeleteFile(string filename, bool confirm)
        {
            DeleteFile(filename, confirm, false);
        }
        public static void DeleteFile(string filename, bool confirm, bool delete_permanently)
        {
            UIOption ui_option = UIOption.OnlyErrorDialogs;
            if (confirm) ui_option = UIOption.AllDialogs;

            RecycleOption recycle_option =
                recycle_option = RecycleOption.SendToRecycleBin;
            if (delete_permanently)
                recycle_option = RecycleOption.DeletePermanently;

            try
            {
                FileSystem.DeleteFile(filename, ui_option, recycle_option);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error deleting file.\n" + ex.Message,
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // If the user checks or unchecks a size node,
        // check or uncheck its file nodes.
        private void trvFiles_AfterCheck(object sender, TreeViewEventArgs e)
        {
            // If this is a size node, check or uncheck its file nodes.
            if (e.Node.Level == 0)
            {
                foreach (TreeNode file_node in e.Node.Nodes)
                {
                    file_node.Checked = e.Node.Checked;
                }
            }
        }

        // The cryptographic service provider.
        private MD5 Md5 = MD5.Create();

        // Compute the file's hash.
        private byte[] GetHash(string filename)
        {
            using (FileStream stream = File.OpenRead(filename))
            {
                return Md5.ComputeHash(stream);
            }
        }

        // Return a hash code as a string.
        private string GetHashString(string filename)
        {
            byte[] bytes = GetHash(filename);
            string result = "";
            foreach (byte b in bytes) result += b.ToString("x2");
            return result;
        }

        // Return a hash code as a string.
        private string BytesToString(byte[] bytes)
        {
            string result = "";
            foreach (byte b in bytes) result += b.ToString("x2");
            return result;
        }

        // Load a bitmap without locking it.
        private Bitmap LoadBitmapUnlocked(string file_name)
        {
            using (Bitmap bm = new Bitmap(file_name))
            {
                return new Bitmap(bm);
            }
        }
    }
}
